/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid Rate application
 * 
 * Copyright (C) 2004, Peter Nixon
 *
 * Peter Nixon <codemonkey@peternixon.net>
 *
 * $Id: app_prepaid_auth_pin.c,v 1.7 2004/09/27 20:04:12 peternixon Exp $
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 *
 * FIXME: Not working yet!
 * 
 */

#define sqlcommandsize 2048

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/utils.h>

#include <asterisk/app.h>
#include <asterisk/say.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>

#include <asterisk/config.h>
#include <pgsql/libpq-fe.h>

#include <asterisk/translate.h>
#include <asterisk/musiconhold.h>
#include <asterisk/callerid.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <asterisk/options.h>


static char *tdesc =
" PrepaidAuthPIN(): Returns the Tariff rate"
" for the destination number.\n";
static char *app = "PrepaidAuthPIN";
static char *config = "prepaid.conf";
static char *synopsis = 
"Prepaid PIN Authentication Application";


static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *psqlauthentication_pin = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int sqlauthentication_pin_alloc = 0;
static int connected = 0; /* Initially not connected to the DB */

PGconn *conn;
ast_mutex_t postgreslock;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

struct calldata {
    char username[256];
    char cardnumber[256];
    char destination[256];
    char *credit;
    char *tariff;
    char *active;
    char *rate;
    char *tech;
    char *prefix;
    char *hostname;
    int timeout;
    char newdestination[256];
};

/* Checks if we are connected to the db - if not then it tries to reconnect */
static int db_check_connected(void)
{
	int i;
	PGresult *result;

	if (connected) {
                /* seems that we are connected - but let us check that */
                for (i=0; i<3; i++) {
                        /* PQstatus is not working as expected - so we are trying a query before PQstatus - hope that helps */
                        result = PQexec(conn, "SELECT 1");
                        PQclear(result);
                        if (PQstatus(conn) == CONNECTION_BAD) {
                                /* seems that we are having troubles with the connection - trying to reset the connection */
                                /* and set connected to false */
                                connected = 0;
                                ast_log(LOG_ERROR, "PrepaidAuthPIN: connection to database server '%s' is broken - I will try to reconnect (up to) 3 times\n", pghostname);
                                PQreset(conn);
                                /* retry it up to 3 times */
                                continue;
                        } else {
                                /* Everything is fine - so go on */
                                connected = 1;
                                break;
                        }
                }
        } else {
		/* We are not connected - so make the initial connection */
                conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
                if (PQstatus(conn) != CONNECTION_BAD) {
                        ast_log(LOG_DEBUG, "PrepaidAuthPIN: Successfully connected to database '%s'.\n", pgdbname);
                        connected = 1;
                } else {
                        ast_log(LOG_ERROR, "PrepaidAuthPIN: FATAL - Couldn't connect to database server '%s'. \n", pghostname);
                        connected = 0;
                }
        }
        return connected;
}

static int prepaidrate_checkcid(struct ast_channel *chan, struct calldata *cd)
{
	char sqlcmd[sqlcommandsize] = "";
	PGresult *result;

	if (db_check_connected()) {
		/* lookup callerid */
	        sprintf(sqlcmd, psqlauthentication_pin, chan->cid.cid_num);
        	ast_log(LOG_DEBUG,"PrepaidAuthPIN: SQL command as follows:  %s\n", sqlcmd);
	        ast_mutex_lock(&postgreslock);
        	result = PQexec(conn, sqlcmd);

		/* Check for Errors */
		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_DEBUG,"PrepaidAuthPIN: FATAL ERROR: %s\n", PQresultErrorMessage(result));
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }
        	if(PQntuples(result)<=0) {
			chan->priority+=100;
			ast_log(LOG_DEBUG,"PrepaidAuthPIN: WARNING: CallerID '%s' is not known. Jumping to Priority '%i'\n", chan->cid.cid_num, chan->priority);
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return 0;
	        }

		strncpy(cd->username, PQgetvalue(result,0,0), sizeof(cd->username));
		ast_log(LOG_DEBUG,"PrepaidAuthPIN: CallerID '%s' is owned by Username '%s'\n", chan->cid.cid_num, cd->username);

		cd->active = PQgetvalue(result,0,1);
		ast_log(LOG_DEBUG,"PrepaidAuthPIN: CallerID '%s' activated='%s'\n", chan->cid.cid_num, cd->active);

	        /* check if the callerid is activated */
        	if(strcmp(cd->active, "t") != 0 ) {
			ast_mutex_unlock(&postgreslock);
			ast_log(LOG_DEBUG,"PrepaidAuthPIN: WARNING: CallerID '%s' for Username '%s' is disabled\n", chan->cid.cid_num, cd->username);
			return -2;
        	}

		PQclear(result);
	        ast_mutex_unlock(&postgreslock);

		return 0;

	} else {
		/* We are not connected to the DB */
		return -1;
	}
}


static int prepaidauthpin_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	struct calldata cd;

	if (!chan->cid.cid_num) {
		chan->priority+=100;
		ast_log (LOG_DEBUG, "PrepaidAuthPIN: No PIN. Jumping to Priority '%i'\n", chan->priority);
		return res;
	} else {
		ast_verbose (VERBOSE_PREFIX_3 "PrepaidAuthPIN: PIN '%s' present\n", chan->cid.cid_num);
	}

	LOCAL_USER_ADD(u);

	res = prepaidrate_checkcid(chan, &cd);
	if (!res) {
                ast_cdr_reset(chan->cdr,0);

		if (cd.username && !ast_strlen_zero(cd.username)) {
			/* Everything seems ok, so set ACCOUNTCODE */
			ast_cdr_setaccount(chan, cd.username);
                	ast_log(LOG_DEBUG,"PrepaidAuthPIN: Setting ACCOUNTCODE to '%s'\n", cd.username);
		}
	} else {
		ast_log(LOG_DEBUG, "PrepaidAuthPIN: No ACCOUNTCODE so exiting now...\n");
		return -1;
	}

	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
        PQfinish(conn);
        connected = 0;
        if (pghostname && hostname_alloc) {
                free(pghostname);
                pghostname = NULL;
                hostname_alloc = 0;
        }
        if (pgdbname && dbname_alloc) {
                free(pgdbname);
                pgdbname = NULL;
                dbname_alloc = 0;
        }
        if (pgdbuser && dbuser_alloc) {
                free(pgdbuser);
                pgdbuser = NULL;
                dbuser_alloc = 0;
        }
        if (pgdbsock && dbsock_alloc) {
                free(pgdbsock);
                pgdbsock = NULL;
                dbsock_alloc = 0;
        }
        if (pgpassword && password_alloc) {
                free(pgpassword);
                pgpassword = NULL;
                password_alloc = 0;
        }
        if (pgdbport && dbport_alloc) {
                free(pgdbport);
                pgdbport = NULL;
                dbport_alloc = 0;
        }
        if (psqlauthentication_pin && sqlauthentication_pin_alloc) {
                free(psqlauthentication_pin);
                psqlauthentication_pin = NULL;
                sqlauthentication_pin_alloc = 0;
        }

	return ast_unregister_application(app);
}

int load_module(void)
{

        int res;
        struct ast_config *cfg;
        struct ast_variable *var;
        char *tmp;

        cfg = ast_config_load(config);
        if (!cfg) {
                ast_log(LOG_WARNING, "Unable to load config for Prepaid Application: %s\n", config);
                return 0;
        }

        var = ast_variable_browse(cfg, "global");
        if (!var) {
                /* nothing configured */
                return 0;
        }

        tmp = ast_variable_retrieve(cfg,"global","hostname");
        if (tmp) {
                pghostname = malloc(strlen(tmp) + 1);
                if (pghostname != NULL) {
                        hostname_alloc = 1;
                        strcpy(pghostname, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database server hostname not specified.  Assuming localhost\n");
                pghostname = "localhost";
        }

        tmp = ast_variable_retrieve(cfg, "global", "dbname");
        if (tmp) {
                pgdbname = malloc(strlen(tmp) + 1);
                if (pgdbname != NULL) {
                        dbname_alloc = 1;
                        strcpy(pgdbname, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database not specified.  Assuming asterisk\n");
                pgdbname = "asterisk";
        }

        tmp = ast_variable_retrieve(cfg, "global", "user");
        if (tmp) {
                pgdbuser = malloc(strlen(tmp) + 1);
                if (pgdbuser != NULL) {
                        dbuser_alloc = 1;
                        strcpy(pgdbuser, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database user not specified.  Assuming postgres\n");
                pgdbuser = "postgres";
        }

        tmp = ast_variable_retrieve(cfg, "global", "password");
        if (tmp) {
                pgpassword = malloc(strlen(tmp) + 1);
                if (pgpassword != NULL) {
                        password_alloc = 1;
                        strcpy(pgpassword, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database password not specified.  Assuming blank\n");
                pgpassword = "";
        }

        tmp = ast_variable_retrieve(cfg,"global","port");
        if (tmp) {
                pgdbport = malloc(strlen(tmp) + 1);
                if (pgdbport != NULL) {
                        dbport_alloc = 1;
                        strcpy(pgdbport, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database port not specified.  Using 5432.\n");
                pgdbport = "5432";
        }

        tmp = ast_variable_retrieve(cfg,"sql","authentication_cid");
        if (tmp) {
                psqlauthentication_pin = malloc(strlen(tmp) + 1);
                if (psqlauthentication_pin != NULL) {
                        sqlauthentication_pin_alloc = 1;
                        strncpy(psqlauthentication_pin,tmp,strlen(tmp));
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid authentication_cid query not specified. Using default.\n");
                psqlauthentication_pin = "SELECT * FROM asterisk_authenticate_cid('%s') AS authenticate_cid(cid text, activated bool);";
        }

        ast_config_destroy(cfg);

	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got hostname of %s\n", pghostname);
	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got port of %s\n", pgdbport);
	if (pgdbsock) {
		ast_log(LOG_DEBUG,"PrepaidAuthPIN: got sock file of %s\n", pgdbsock);
	}
	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got user of %s\n", pgdbuser);
	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got dbname of %s\n", pgdbname);
	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got password of %s\n", pgpassword);

	ast_log(LOG_DEBUG,"PrepaidAuthPIN: got sqlauthentication_pin of '%s'\n", psqlauthentication_pin);

        /* Check DB connection (it will automatically set up the connection) */
	db_check_connected();
	/* Initialise a mutex to make sure Postgres doesn't walk all over itself */
	ast_mutex_init(&postgreslock);

	res = ast_register_application(app, prepaidauthpin_exec, synopsis, tdesc);
	if (res) {
		ast_log(LOG_ERROR, "Unable to register PrepaidAuthPIN Application\n");
	}

	return res;

}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
