/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid CDR logger
 *
 * Copyright (C) 2004, Peter Nixon
 *
 * Peter Nixon <codemonkey@peternixon.net>
 *
 * Adapted from the Postgresql CDR logger originally by Matthew D. Hardeman
 * Adapted from the MySQL CDR logger originally by James Sharp 
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 * $Id: cdr_prepaid.c,v 1.10 2004/10/01 13:42:08 tuyano Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include <sys/types.h>
#include <asterisk/config.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/cdr.h>
#include <asterisk/module.h>
#include <asterisk/pbx.h>
#include <asterisk/logger.h>
#include <asterisk/asterisk.h>

#include <pgsql/libpq-fe.h>

#define DATE_FORMAT "%Y-%m-%d %T"

static char *desc = "Prepaid CDR Backend";
static char *name = "cdrprepaid";
static char *config = "prepaid.conf";
static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *chan_rate = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int chrate_alloc = 0;
static int connected = 0;

AST_MUTEX_DEFINE_STATIC(pgsql_lock);

PGconn		*conn;
PGresult	*result;

static int pgsql_log(struct ast_cdr *cdr)
{
	struct tm tm;
	char sqlcmd[2048] = "", timestr[128];
	char *pgerror;
	struct ast_channel *chan;

	if ((chan_rate=(char*) malloc(sizeof(char)*5))!=NULL)
			chrate_alloc=1;			

	ast_mutex_lock(&pgsql_lock);

	localtime_r(&cdr->start.tv_sec,&tm);
	strftime(timestr, sizeof(timestr), DATE_FORMAT, &tm);

	if ((!connected) && pghostname && pgdbuser && pgpassword && pgdbname) {
		conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
		if (PQstatus(conn) != CONNECTION_BAD) {
			connected = 1;
		} else {
			pgerror = PQerrorMessage(conn);
			ast_log(LOG_ERROR, "cdr_prepaid: Unable to connect to database server %s.  Calls will not be logged!\n", pghostname);
                        ast_log(LOG_ERROR, "cdr_prepaid: Reason: %s\n", pgerror);
		}
	}

	if (connected) {
		char *clid=NULL, *dcontext=NULL, *channel=NULL, *dstchannel=NULL, *lastapp=NULL, *lastdata=NULL;
		char *uniqueid=NULL, *userfield=NULL;
                char hostname[256];


                if (gethostname(hostname, sizeof(hostname)))
                     strncpy(hostname, "<Unknown>", sizeof(hostname)-1);


		/* Maximum space needed would be if all characters needed to be escaped, plus a trailing NULL */
		if ((clid = alloca(strlen(cdr->clid) * 2 + 1)) != NULL)
			PQescapeString(clid, cdr->clid, strlen(cdr->clid));
		if ((dcontext = alloca(strlen(cdr->dcontext) * 2 + 1)) != NULL)
			PQescapeString(dcontext, cdr->dcontext, strlen(cdr->dcontext));
		if ((channel = alloca(strlen(cdr->channel) * 2 + 1)) != NULL)
			PQescapeString(channel, cdr->channel, strlen(cdr->channel));
		if ((dstchannel = alloca(strlen(cdr->dstchannel) * 2 + 1)) != NULL)
			PQescapeString(dstchannel, cdr->dstchannel, strlen(cdr->dstchannel));
		if ((lastapp = alloca(strlen(cdr->lastapp) * 2 + 1)) != NULL)
			PQescapeString(lastapp, cdr->lastapp, strlen(cdr->lastapp));
		if ((lastdata = alloca(strlen(cdr->lastdata) * 2 + 1)) != NULL)
			PQescapeString(lastdata, cdr->lastdata, strlen(cdr->lastdata));
		if ((uniqueid = alloca(strlen(cdr->uniqueid) * 2 + 1)) != NULL)
			PQescapeString(uniqueid, cdr->uniqueid, strlen(cdr->uniqueid));
		if ((userfield = alloca(strlen(cdr->userfield) * 2 + 1)) != NULL)
			PQescapeString(userfield, cdr->userfield, strlen(cdr->userfield));

		/* Check for all alloca failures above at once */
		if ((!clid) || (!dcontext) || (!channel) || (!dstchannel) || (!lastapp) || (!lastdata) || (!uniqueid) || (!userfield)) {
			ast_log(LOG_ERROR, "cdr_prepaid:  Out of memory error (insert fails)\n");
			ast_mutex_unlock(&pgsql_lock);
			return -1;
		}
		chan = ast_get_channel_by_name_locked(cdr->channel);
		chan_rate=pbx_builtin_getvar_helper(chan,"RATE");
		if (!chan_rate){
                  ast_log(LOG_DEBUG,"No rate to destination, lets take it to be 1\n");
		  chan_rate=(char*) malloc(5);
		  strncpy(chan_rate,"1",sizeof(char) * 2);
		}	
		ast_log(LOG_DEBUG,"cdr_prepaid: inserting a CDR record with rate %s.\n",chan_rate);

		snprintf(sqlcmd,sizeof(sqlcmd),"INSERT INTO call (uniqueid,sessionid,username,nasipaddress,stoptime,sessiontime,calledstation,terminatecause,calledrate) VALUES ('%s','%s','%s','%s','NOW()','%i','%s','%s',%i)",cdr->uniqueid,cdr->uniqueid,chan->cid.cid_num,hostname,cdr->billsec,cdr->dst,cdr->lastapp,atoi(chan_rate));
		ast_log(LOG_DEBUG,"cdr_prepaid: SQL command executed:  %s\n",sqlcmd);
	
                //ast_mutex_unlock(&chan->lock);

		/* Test to be sure we're still connected... */
		/* If we're connected, and connection is working, good. */
		/* Otherwise, attempt reconnect.  If it fails... sorry... */
		if (PQstatus(conn) == CONNECTION_OK) {
			connected = 1;
		} else {
			ast_log(LOG_ERROR, "cdr_prepaid: Connection was lost... attempting to reconnect.\n");
			PQreset(conn);
			if (PQstatus(conn) == CONNECTION_OK) {
				ast_log(LOG_ERROR, "cdr_prepaid: Connection reestablished.\n");
				connected = 1;
			} else {
				pgerror = PQerrorMessage(conn);
				ast_log(LOG_ERROR, "cdr_prepaid: Unable to reconnect to database server %s. Calls will not be logged!\n", pghostname);
				ast_log(LOG_ERROR, "cdr_prepaid: Reason: %s\n", pgerror);
				connected = 0;
				ast_mutex_unlock(&pgsql_lock);
				return -1;
			}
		}
		result = PQexec(conn, sqlcmd);
		if ( PQresultStatus(result) != PGRES_COMMAND_OK) {
                        pgerror = PQresultErrorMessage(result);
			ast_log(LOG_ERROR,"cdr_prepaid: Failed to insert call detail record into database!\n");
                        ast_log(LOG_ERROR,"cdr_prepaid: Reason: %s\n", pgerror);
			ast_mutex_unlock(&pgsql_lock);
			return -1;
		}
	}
	ast_mutex_unlock(&pgsql_lock);
	return 0;
}

char *description(void)
{
	return desc;
}

static int my_unload_module(void)
{ 
	if (conn)
		PQfinish(conn);
	conn = NULL;
	connected = 0;
	if (pghostname && hostname_alloc) {
		free(pghostname);
		pghostname = NULL;
		hostname_alloc = 0;
	}
	if (pgdbname && dbname_alloc) {
		free(pgdbname);
		pgdbname = NULL;
		dbname_alloc = 0;
	}
	if (pgdbuser && dbuser_alloc) {
		free(pgdbuser);
		pgdbuser = NULL;
		dbuser_alloc = 0;
	}
	if (pgdbsock && dbsock_alloc) {
		free(pgdbsock);
		pgdbsock = NULL;
		dbsock_alloc = 0;
	}
	if (pgpassword && password_alloc) {
		free(pgpassword);
		pgpassword = NULL;
		password_alloc = 0;
	}
	if (pgdbport && dbport_alloc) {
		free(pgdbport);
		pgdbport = NULL;
		dbport_alloc = 0;
	}
        /*
	if (chan_rate && chrate_alloc) {
		free(chan_rate);
		chan_rate = NULL;
		chrate_alloc = 0;
	}
        */
	ast_cdr_unregister(name);
	return 0;
}

static int my_load_module(void)
{
	int res;
	struct ast_config *cfg;
	struct ast_variable *var;
        char *pgerror;
	char *tmp;

	cfg = ast_config_load(config);
	if (!cfg) {
		ast_log(LOG_WARNING, "cdr_prepaid: Unable to load config for PostgreSQL CDR's: %s\n", config);
		return 0;
	}
	
	var = ast_variable_browse(cfg, "global");
	if (!var) {
		/* nothing configured */
		return 0;
	}

	tmp = ast_variable_retrieve(cfg,"global","hostname");
	if (tmp) {
		pghostname = malloc(strlen(tmp) + 1);
		if (pghostname != NULL) {
			memset(pghostname, 0, strlen(tmp) + 1);
			hostname_alloc = 1;
			strncpy(pghostname, tmp, strlen(tmp));
		} else {
			ast_log(LOG_ERROR,"cdr_prepaid: Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"cdr_prepaid: PostgreSQL server hostname not specified.  Assuming localhost\n");
		pghostname = "localhost";
	}

	tmp = ast_variable_retrieve(cfg,"global","dbname");
	if (tmp) {
		pgdbname = malloc(strlen(tmp) + 1);
		if (pgdbname != NULL) {
			memset(pgdbname, 0, strlen(tmp) + 1);
			dbname_alloc = 1;
			strncpy(pgdbname, tmp, strlen(tmp));
		} else {
			ast_log(LOG_ERROR,"cdr_prepaid: Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"cdr_prepaid: PostgreSQL database not specified.  Assuming asterisk\n");
		pgdbname = "asterisk";
	}

	tmp = ast_variable_retrieve(cfg,"global","user");
	if (tmp) {
		pgdbuser = malloc(strlen(tmp) + 1);
		if (pgdbuser != NULL) {
			memset(pgdbuser, 0, strlen(tmp) + 1);
			dbuser_alloc = 1;
			strncpy(pgdbuser, tmp, strlen(tmp));
		} else {
			ast_log(LOG_ERROR,"cdr_prepaid: Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"cdr_prepaid: PostgreSQL database user not specified.  Assuming root\n");
		pgdbuser = "root";
	}

	tmp = ast_variable_retrieve(cfg,"global","password");
	if (tmp) {
		pgpassword = malloc(strlen(tmp) + 1);
		if (pgpassword != NULL) {
			memset(pgpassword, 0, strlen(tmp) + 1);
			password_alloc = 1;
			strncpy(pgpassword, tmp, strlen(tmp));
		} else {
			ast_log(LOG_ERROR,"cdr_prepaid: Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"cdr_prepaid: PostgreSQL database password not specified.  Assuming blank\n");
		pgpassword = "";
	}

	tmp = ast_variable_retrieve(cfg,"global","port");
	if (tmp) {
		pgdbport = malloc(strlen(tmp) + 1);
		if (pgdbport != NULL) {
			memset(pgdbport, 0, strlen(tmp) + 1);
			dbport_alloc = 1;
			strncpy(pgdbport, tmp, strlen(tmp));
		} else {
			ast_log(LOG_ERROR,"cdr_prepaid: Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"cdr_prepaid: PostgreSQL database port not specified.  Using default 5432.\n");
		pgdbport = "5432";
	}

	ast_config_destroy(cfg);

	ast_log(LOG_DEBUG,"cdr_prepaid: got hostname of %s\n",pghostname);
	ast_log(LOG_DEBUG,"cdr_prepaid: got port of %s\n",pgdbport);
	if (pgdbsock)
		ast_log(LOG_DEBUG,"cdr_prepaid: got sock file of %s\n",pgdbsock);
	ast_log(LOG_DEBUG,"cdr_prepaid: got user of %s\n",pgdbuser);
	ast_log(LOG_DEBUG,"cdr_prepaid: got dbname of %s\n",pgdbname);
	ast_log(LOG_DEBUG,"cdr_prepaid: got password of %s\n",pgpassword);

	conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
	if (PQstatus(conn) != CONNECTION_BAD) {
		ast_log(LOG_DEBUG,"cdr_prepaid: Successfully connected to PostgreSQL database.\n");
		connected = 1;
	} else {
                pgerror = PQerrorMessage(conn);
		ast_log(LOG_ERROR, "cdr_prepaid: Unable to connect to database server %s.  Calls will not be logged!\n", pghostname);
                ast_log(LOG_ERROR, "cdr_prepaid: Reason: %s\n", pgerror);
		connected = 0;
	}

	res = ast_cdr_register(name, desc, pgsql_log);
	if (res) {
		ast_log(LOG_ERROR, "Unable to register PGSQL CDR handling\n");
	}
	return res;
}

int load_module(void)
{
	return my_load_module();
}

int unload_module(void)
{
	return my_unload_module();
}

int reload(void)
{
	my_unload_module();
	return my_load_module();
}

int usecount(void)
{
	return connected;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
